@@ -361,6 +361,7 @@ function filter_end_date() {{
         );
     }
 
+    md += "\n";
     md += &pager_md;
     Ok(response::markdown(&md))
 }
@@ -484,5 +485,328 @@ pub async fn get_execution(
         );
     }
 
+    md += &format!("- [**Graph**](/executions/{}/graph)\n", execution.id);
+
     Ok(response::markdown(&md))
 }
+
+#[get("/executions/{id}/graph")]
+pub async fn get_execution_graph(
+    config: web::Data<Config>,
+    path: web::Path<(i32,)>,
+) -> actix_web::Result<HttpResponse> {
+    let id = path.0;
+    let mut store = config.connect_metadata_store().await?;
+
+    let graph = Graph::new(&mut store, id)
+        .await
+        .map_err(actix_web::error::ErrorInternalServerError)?;
+
+    let svg = std::process::Command::new("dot")
+        .arg("-Tsvg")
+        .stdin(std::process::Stdio::piped())
+        .stdout(std::process::Stdio::piped())
+        .spawn()
+        .ok()
+        .and_then(|mut child| {
+            use std::io::Write;
+
+            let writer = child.stdin.as_mut()?;
+            graph.render(writer).ok()?;
+            writer.flush().ok()?;
+            let output = child.wait_with_output().ok()?;
+            if !output.status.success() {
+                None
+            } else {
+                Some(output.stdout)
+            }
+        });
+
+    if let Some(svg) = svg {
+        Ok(response::svg(&String::from_utf8(svg).expect("TODO")))
+    } else {
+        let mut buf = Vec::new();
+        graph.render(&mut buf).expect("TODO");
+        Ok(response::markdown(&String::from_utf8(buf).expect("TODO")))
+    }
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+pub enum NodeId {
+    Execution(i32),
+    Artifact(i32),
+}
+
+impl std::fmt::Display for NodeId {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            Self::Execution(x) => write!(f, "E{}", x),
+            Self::Artifact(x) => write!(f, "A{}", x),
+        }
+    }
+}
+
+#[derive(Debug, Clone)]
+pub enum Node {
+    Execution {
+        node: Execution,
+        inputs: usize,
+        outputs: usize,
+    },
+    Artifact {
+        node: crate::mlmd::artifact::Artifact,
+        inputs: usize,
+        outputs: usize,
+    },
+}
+
+impl Node {
+    pub fn id(&self) -> NodeId {
+        match self {
+            Self::Execution { node, .. } => NodeId::Execution(node.id),
+            Self::Artifact { node, .. } => NodeId::Artifact(node.id),
+        }
+    }
+
+    pub fn set_in_out(&mut self, events: &[mlmd::metadata::Event]) {
+        use mlmd::metadata::EventType::*;
+
+        let mut n_input = 0;
+        let mut n_output = 0;
+        for event in events {
+            match event.ty {
+                Input | DeclaredInput | InternalInput => {
+                    n_input += 1;
+                }
+                Output | DeclaredOutput | InternalOutput => {
+                    n_output += 1;
+                }
+                _ => {}
+            }
+        }
+        match self {
+            Self::Execution {
+                inputs, outputs, ..
+            } => {
+                *inputs = n_input;
+                *outputs = n_output;
+            }
+            Self::Artifact {
+                inputs, outputs, ..
+            } => {
+                *inputs = n_input;
+                *outputs = n_output;
+            }
+        }
+    }
+
+    pub fn label(&self) -> String {
+        match self {
+            Self::Execution {
+                node,
+                inputs,
+                outputs,
+            } => format!(
+                "{}\n{}\nin={},out={}",
+                node.id, node.type_name, inputs, outputs
+            ),
+            Self::Artifact {
+                node,
+                inputs,
+                outputs,
+            } => format!(
+                "{}\n{}\nin={},out={}",
+                node.id, node.type_name, inputs, outputs
+            ),
+        }
+    }
+
+    pub fn url(&self) -> String {
+        match self {
+            Self::Execution { node, .. } => format!("/executions/{}", node.id),
+            Self::Artifact { node, .. } => format!("/artifacts/{}", node.id),
+        }
+    }
+
+    pub fn shape(&self) -> String {
+        match self {
+            Self::Execution { .. } => format!("box"),
+            Self::Artifact { .. } => format!("ellipse"),
+        }
+    }
+
+    pub fn attrs(&self) -> Vec<String> {
+        vec![
+            format!("label={:?}", self.label()),
+            format!("shape={:?}", self.shape()),
+            format!("URL={:?}", self.url()),
+        ]
+    }
+}
+
+#[derive(Debug, Clone)]
+pub struct Edge {
+    pub source: NodeId,
+    pub target: NodeId,
+    pub event: crate::mlmd::event::Event,
+}
+
+#[derive(Debug)]
+struct Graph {
+    nodes: Vec<Node>,
+    edges: Vec<Edge>,
+}
+
+impl Graph {
+    async fn new(store: &mut mlmd::MetadataStore, execution_id: i32) -> anyhow::Result<Self> {
+        let mut nodes = HashMap::new();
+        let mut edges = Vec::new();
+        let mut stack = vec![NodeId::Execution(execution_id)];
+        while let Some(curr) = stack.pop() {
+            if nodes.contains_key(&curr) {
+                continue;
+            }
+            let mut curr = match curr {
+                NodeId::Execution(id) => Node::Execution {
+                    node: fetch_execution(store, id).await?,
+                    inputs: 0,
+                    outputs: 0,
+                },
+                NodeId::Artifact(id) => Node::Artifact {
+                    node: fetch_artifact(store, id).await?,
+                    inputs: 0,
+                    outputs: 0,
+                },
+            };
+
+            let events = match &curr {
+                Node::Execution { node, .. } => {
+                    store
+                        .get_events()
+                        .execution(mlmd::metadata::ExecutionId::new(node.id))
+                        .execute()
+                        .await?
+                }
+                Node::Artifact { node, .. } => {
+                    store
+                        .get_events()
+                        .artifact(mlmd::metadata::ArtifactId::new(node.id))
+                        .execute()
+                        .await?
+                }
+            };
+            curr.set_in_out(&events);
+            nodes.insert(curr.id(), curr.clone());
+            anyhow::ensure!(
+                nodes.len() < 100,
+                "Too many executions and artifact to visualize"
+            );
+
+            for event in events {
+                if matches!(curr, Node::Artifact { .. }) {
+                    use mlmd::metadata::EventType::*;
+                    if matches!(event.ty, Output | DeclaredOutput | InternalOutput) {
+                        let id = NodeId::Execution(event.execution_id.get());
+                        stack.push(id);
+                        edges.push(Edge {
+                            source: id,
+                            target: curr.id(),
+                            event: event.into(),
+                        });
+                    }
+                } else {
+                    use mlmd::metadata::EventType::*;
+                    if event.execution_id.get() == execution_id
+                        || matches!(event.ty, Input | DeclaredInput | InternalInput)
+                    {
+                        let id = NodeId::Artifact(event.artifact_id.get());
+                        stack.push(id);
+                        if matches!(event.ty, Input | DeclaredInput | InternalInput) {
+                            edges.push(Edge {
+                                source: id,
+                                target: curr.id(),
+                                event: event.into(),
+                            });
+                        }
+                    }
+                }
+            }
+        }
+
+        Ok(Self {
+            nodes: nodes.into_iter().map(|x| x.1).collect(),
+            edges,
+        })
+    }
+
+    fn render<W: std::io::Write>(&self, writer: &mut W) -> anyhow::Result<()> {
+        writeln!(writer, "digraph execution_graph {{")?;
+
+        for node in &self.nodes {
+            writeln!(writer, "{}[{}]", node.id(), node.attrs().join(","))?;
+        }
+
+        for edge in &self.edges {
+            writeln!(
+                writer,
+                "{} -> {} [label={:?}];",
+                edge.source,
+                edge.target,
+                format!("{:?}:{:?}", edge.event.ty, edge.event.path)
+            )?;
+        }
+
+        writeln!(writer, "}}")?;
+        Ok(())
+    }
+}
+
+async fn fetch_execution(store: &mut mlmd::MetadataStore, id: i32) -> anyhow::Result<Execution> {
+    let executions = store
+        .get_executions()
+        .id(mlmd::metadata::ExecutionId::new(id))
+        .execute()
+        .await?;
+    anyhow::ensure!(!executions.is_empty(), "no such execution: {}", id);
+
+    let types = store
+        .get_execution_types()
+        .id(executions[0].type_id)
+        .execute()
+        .await?;
+    anyhow::ensure!(
+        !executions.is_empty(),
+        "no such execution tyep: {}",
+        executions[0].type_id.get()
+    );
+
+    Ok(Execution::from((types[0].clone(), executions[0].clone())))
+}
+
+async fn fetch_artifact(
+    store: &mut mlmd::MetadataStore,
+    id: i32,
+) -> anyhow::Result<crate::mlmd::artifact::Artifact> {
+    let artifacts = store
+        .get_artifacts()
+        .id(mlmd::metadata::ArtifactId::new(id))
+        .execute()
+        .await?;
+    anyhow::ensure!(!artifacts.is_empty(), "no such artifact: {}", id);
+
+    let types = store
+        .get_artifact_types()
+        .id(artifacts[0].type_id)
+        .execute()
+        .await?;
+    anyhow::ensure!(
+        !artifacts.is_empty(),
+        "no such artifact tyep: {}",
+        artifacts[0].type_id.get()
+    );
+
+    Ok(crate::mlmd::artifact::Artifact::from((
+        types[0].clone(),
+        artifacts[0].clone(),
+    )))
+}
